#ifndef _WDL_DIRSCAN_H_
#define _WDL_DIRSCAN_H_

#include "string.h"

class WDL_DirScan
{
  public:
    WDL_DirScan() : m_h(INVALID_HANDLE_VALUE)
    {
    }

    ~WDL_DirScan()
    {
      Close();
    }

    int First(const char *dirname) // returns 0 if success
    {
      int l=strlen(dirname);
      if (l < 1) return -1;

      WDL_String scanstr(dirname);
      if (dirname[l-1] == '\\' || dirname[l-1] == '/') scanstr.SetLen(l-1);
      m_leading_path.Set(scanstr.Get());
      scanstr.Append("\\*");

      Close();
      m_h=FindFirstFile(scanstr.Get(),&m_fd);
      return (m_h == INVALID_HANDLE_VALUE);
    }
    int Next() // returns 0 on success
    {
      if (m_h == INVALID_HANDLE_VALUE) return -1;
      return !FindNextFile(m_h,&m_fd);
    }
    void Close()
    {
      if (m_h != INVALID_HANDLE_VALUE) FindClose(m_h);
      m_h=INVALID_HANDLE_VALUE;
    }

    char *GetCurrentFN() { return m_fd.cFileName; }
    void GetCurrentFullFN(WDL_String *str) 
    { 
      str->Set(m_leading_path.Get()); 
      str->Append("\\"); 
      str->Append(m_fd.cFileName); 
    }
    int GetCurrentIsDirectory() { return !!(m_fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY); }

    // these are somewhat windows specific calls, eh
    DWORD GetCurrentFileSize(DWORD *HighWord=NULL) { if (HighWord) *HighWord = m_fd.nFileSizeHigh; return m_fd.nFileSizeLow; }
    void GetCurrentLastWriteTime(FILETIME *ft) { *ft = m_fd.ftLastWriteTime; }
    void GetCurrentLastAccessTime(FILETIME *ft) { *ft = m_fd.ftLastAccessTime; }
    void GetCurrentCreationTime(FILETIME *ft) { *ft = m_fd.ftCreationTime; }

  private:
    WIN32_FIND_DATA m_fd;
    HANDLE m_h;
    WDL_String m_leading_path;
};

#endif